/**
 * Xlib shim to monitor pixmap creates/destroys.  When too much pixmap memory
 * is used, exit.
 *
 * Brian Paul
 * 19 Sep 2007
 * 
 */


#define MEMORY_LIMIT 200000  /* bytes */


#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#define __USE_GNU
#include <dlfcn.h>


struct pixmap_node
{
   Pixmap Id;
   unsigned Bytes;
   struct pixmap_node *Next;
};


struct wrapped
{
   Display *(*XOpenDisplay_func)(_Xconst char *name);
   int (*XCloseDisplay_func)(Display *dpy);
   Pixmap (*XCreatePixmap_func)(Display *display, Drawable d, unsigned int width,
                                unsigned int height, unsigned int depth);

   int (*XFreePixmap_func)(Display *display, Pixmap pixmap);
};


struct shim_info
{
   void *XLibHandle;
   struct wrapped Wrapped;
   unsigned BytesAllocated;
   struct pixmap_node *List;
};

static struct shim_info *Shim = NULL;


static struct shim_info *
InitShim(void)
{
   const char *xlibname = "/usr/lib/libX11.so";
   const int mode = RTLD_LAZY;

   static struct shim_info *shim = NULL;

   if (!shim) {
      shim = (struct shim_info *) calloc(1, sizeof(struct shim_info));

      /* XXX may don't open libX11.so and use RTLD_DEFAULT handle instead */
#if 0
      libname = RTLD_NEXT;
#endif
      shim->XLibHandle = dlopen(xlibname, mode);
      if (!shim->XLibHandle) {
         fprintf(stderr, "Unable to open %s: %s\n", xlibname, dlerror());
         return NULL;
      }

      shim->Wrapped.XOpenDisplay_func = dlsym(shim->XLibHandle, "XOpenDisplay");
      shim->Wrapped.XCloseDisplay_func = dlsym(shim->XLibHandle, "XCloseDisplay");
      shim->Wrapped.XCreatePixmap_func = dlsym(shim->XLibHandle, "XCreatePixmap");
      shim->Wrapped.XFreePixmap_func = dlsym(shim->XLibHandle, "XFreePixmap");
   }
   return shim;
}


/**
 ** Intercepted functions follow
 **/

Display *
XOpenDisplay(_Xconst char *name)
{
   struct shim_info *shim = InitShim();
   Display *dpy;

   printf("SHIM: XOpenDisplay(%s)\n", name);

   dpy = shim->Wrapped.XOpenDisplay_func(name);
   if (dpy) {
      Shim = shim;
   }

   printf("SHIM: XOpenDisplay(%s) returning %p\n", name, dpy);

   return dpy;
}


int
XCloseDisplay(Display *dpy)
{
   return Shim->Wrapped.XCloseDisplay_func(dpy);
}



Pixmap
XCreatePixmap(Display *display, Drawable d, unsigned int width,
              unsigned int height, unsigned int depth)
{
   struct pixmap_node *node;
   unsigned bytes;
   Pixmap pixmap;

   if (depth >= 8)
      bytes = width * height * ((depth + 7) / 8);
   else
      bytes = height * (width * depth) / 8;

   Shim->BytesAllocated += bytes;

   fprintf(stderr,
           "SHIM: intercepted XCreatePixmap (%u bytes, total = %u bytes)\n",
           bytes, Shim->BytesAllocated);

   if (Shim->BytesAllocated > MEMORY_LIMIT) {
      fprintf(stderr, "SHIM: Pixmap memory limit (%u bytes) hit, exiting!\n",
              MEMORY_LIMIT);
      exit(1);
   }

   pixmap = Shim->Wrapped.XCreatePixmap_func(display, d, width, height, depth);

   /* add pixmap node to list */
   node = (struct pixmap_node *) malloc(sizeof(*node));
   node->Id = pixmap;
   node->Bytes = bytes;
   node->Next = Shim->List;
   Shim->List = node;

   return pixmap;
}


int
XFreePixmap(Display *display, Pixmap pixmap)
{
   struct pixmap_node *prev, *node;
   int found = 0;

   prev = NULL;
   for (node = Shim->List; node; node = node->Next) {
      if (node->Id == pixmap) {
         /* found it */
         fprintf(stderr, "SHIM: intercepted XFreePixmap (%u bytes)\n",
                 node->Bytes);

         Shim->BytesAllocated -= node->Bytes;

         /* remove from list */
         if (prev)
            prev->Next = node->Next;
         else
            Shim->List = node->Next;

         free(node);
         found = 1;
         break;
      }
   }

   if (!found) {
      fprintf(stderr, "SHIM: XFreePixmap failed to find pixmap %u\n",
              (unsigned) pixmap);
   }

   return Shim->Wrapped.XFreePixmap_func(display, pixmap);
}


